@@ -4,9 +4,9 @@ imdb package.
 This package can be used to retrieve information about a movie or
 a person from the IMDb database.
 It can fetch data through different media (e.g.: the IMDb web pages,
-a local installation, a SQL database, etc.)
+a SQL database, etc.)
 
-Copyright 2004-2008 Davide Alberani <da@erlug.linux.it>
+Copyright 2004-2010 Davide Alberani <da@erlug.linux.it>
 
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
@@ -25,18 +25,21 @@ Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
 __all__ = ['IMDb', 'IMDbError', 'Movie', 'Person', 'Character', 'Company',
             'available_access_systems']
-__version__ = VERSION = '3.8'
+__version__ = VERSION = '4.7'
 
 # Import compatibility module (importing it is enough).
 import _compat
 
-import sys, os, ConfigParser
-from types import UnicodeType, TupleType, ListType, MethodType
+import sys, os, ConfigParser, logging
+from types import MethodType
 
 from imdb import Movie, Person, Character, Company
+import imdb._logging
 from imdb._exceptions import IMDbError, IMDbDataAccessError
 from imdb.utils import build_title, build_name, build_company_name
 
+_aux_logger = logging.getLogger('imdbpy.aux')
+
 
 # URLs of the main pages for movies, persons, characters and queries.
 imdbURL_base = 'http://akas.imdb.com/'
@@ -56,6 +59,12 @@ imdbURL_character_main = imdbURL_character_base + 'ch%s/'
 imdbURL_company_base = '%scompany/' % imdbURL_base
 # http://akas.imdb.com/company/co%s/
 imdbURL_company_main = imdbURL_company_base + 'co%s/'
+# http://akas.imdb.com/keyword/%s/
+imdbURL_keyword_main = imdbURL_base + 'keyword/%s/'
+# http://akas.imdb.com/chart/top
+imdbURL_top250 = imdbURL_base + 'chart/top'
+# http://akas.imdb.com/chart/bottom
+imdbURL_bottom100 = imdbURL_base + 'chart/bottom'
 # http://akas.imdb.com/find?%s
 imdbURL_find = imdbURL_base + 'find?%s'
 
@@ -91,8 +100,7 @@ class ConfigParserWithCase(ConfigParser.ConfigParser):
                 self.read(fname)
             except (ConfigParser.MissingSectionHeaderError,
                     ConfigParser.ParsingError), e:
-                import warnings
-                warnings.warn('Troubles reading config file: %s' % e)
+                _aux_logger.warn('Troubles reading config file: %s' % e)
             # Stop at the first valid file.
             if self.has_section('imdbpy'):
                 break
@@ -148,29 +156,41 @@ def IMDb(accessSystem=None, *arguments, **keywords):
             kwds.update(keywords)
             keywords = kwds
         except Exception, e:
-            import warnings
-            warnings.warn('Unable to read configuration file; ' + \
-                            'complete error: %s' % e)
+            logging.getLogger('imdbpy').warn('Unable to read configuration' \
+                                            ' file; complete error: %s' % e)
             # It just LOOKS LIKE a bad habit: we tried to read config
             # options from some files, but something is gone horribly
             # wrong: ignore everything and pretend we were called with
             # the 'http' accessSystem.
             accessSystem = 'http'
+    if 'loggingLevel' in keywords:
+        imdb._logging.setLevel(keywords['loggingLevel'])
+        del keywords['loggingLevel']
+    if 'loggingConfig' in keywords:
+        logCfg = keywords['loggingConfig']
+        del keywords['loggingConfig']
+        try:
+            import logging.config
+            logging.config.fileConfig(os.path.expanduser(logCfg))
+        except Exception, e:
+            logging.getLogger('imdbpy').warn('unable to read logger ' \
+                                            'config: %s' % e)
     if accessSystem in ('http', 'web', 'html'):
         from parser.http import IMDbHTTPAccessSystem
         return IMDbHTTPAccessSystem(*arguments, **keywords)
     elif accessSystem in ('httpThin', 'webThin', 'htmlThin'):
+        import logging
+        logging.warn('httpThin is badly broken and' \
+                                        ' will not be fixed; please switch' \
+                                        ' to "http" or "mobile"')
         from parser.http import IMDbHTTPAccessSystem
         return IMDbHTTPAccessSystem(isThin=1, *arguments, **keywords)
     elif accessSystem in ('mobile',):
         from parser.mobile import IMDbMobileAccessSystem
         return IMDbMobileAccessSystem(*arguments, **keywords)
     elif accessSystem in ('local', 'files'):
-        try:
-            from parser.local import IMDbLocalAccessSystem
-        except ImportError:
-            raise IMDbError, 'the local access system is not installed'
-        return IMDbLocalAccessSystem(*arguments, **keywords)
+        # The local access system was removed since IMDbPY 4.2.
+        raise IMDbError, 'the local access system was removed since IMDbPY 4.2'
     elif accessSystem in ('sql', 'db', 'database'):
         try:
             from parser.sql import IMDbSqlAccessSystem
@@ -196,11 +216,6 @@ def available_access_systems():
         asList.append('mobile')
     except ImportError:
         pass
-    try:
-        from parser.local import IMDbLocalAccessSystem
-        asList.append('local')
-    except ImportError:
-        pass
     try:
         from parser.sql import IMDbSqlAccessSystem
         asList.append('sql')
@@ -228,7 +243,10 @@ class IMDbBase:
     # in the subclasses).
     accessSystem = 'UNKNOWN'
 
-    def __init__(self, defaultModFunct=None, results=20,
+    # Top-level logger for IMDbPY.
+    _imdb_logger = logging.getLogger('imdbpy')
+
+    def __init__(self, defaultModFunct=None, results=20, keywordsResults=100,
                 *arguments, **keywords):
         """Initialize the access system.
         If specified, defaultModFunct is the function used by
@@ -246,6 +264,13 @@ class IMDbBase:
         if results < 1:
             results = 20
         self._results = results
+        try:
+            keywordsResults = int(keywordsResults)
+        except (TypeError, ValueError):
+            keywordsResults = 100
+        if keywordsResults < 1:
+            keywordsResults = 100
+        self._keywordsResults = keywordsResults
 
     def _normalize_movieID(self, movieID):
         """Normalize the given movieID."""
@@ -335,13 +360,15 @@ class IMDbBase:
         self.update(movie, info)
         return movie
 
+    get_episode = get_movie
+
     def _search_movie(self, title, results):
         """Return a list of tuples (movieID, {movieData})"""
         # XXX: for the real implementation, see the method of the
         #      subclass, somewhere under the imdb.parser package.
         raise NotImplementedError, 'override this method'
 
-    def search_movie(self, title, results=None):
+    def search_movie(self, title, results=None, _episodes=False):
         """Return a list of Movie objects for a query for the given title.
         The results argument is the maximum number of results to return."""
         if results is None:
@@ -352,13 +379,28 @@ class IMDbBase:
             results = 20
         # XXX: I suppose it will be much safer if the user provides
         #      an unicode string... this is just a guess.
-        if not isinstance(title, UnicodeType):
+        if not isinstance(title, unicode):
             title = unicode(title, encoding, 'replace')
-        res = self._search_movie(title, results)
+        if not _episodes:
+            res = self._search_movie(title, results)
+        else:
+            res = self._search_episode(title, results)
         return [Movie.Movie(movieID=self._get_real_movieID(mi),
                 data=md, modFunct=self._defModFunct,
                 accessSystem=self.accessSystem) for mi, md in res][:results]
 
+    def _search_episode(self, title, results):
+        """Return a list of tuples (movieID, {movieData})"""
+        # XXX: for the real implementation, see the method of the
+        #      subclass, somewhere under the imdb.parser package.
+        raise NotImplementedError, 'override this method'
+
+    def search_episode(self, title, results=None):
+        """Return a list of Movie objects for a query for the given title.
+        The results argument is the maximum number of results to return;
+        this method searches only for titles of tv (mini) series' episodes."""
+        return self.search_movie(title, results=results, _episodes=True)
+
     def get_person(self, personID, info=Person.Person.default_info,
                     modFunct=None):
         """Return a Person object for the given personID.
@@ -397,7 +439,7 @@ class IMDbBase:
             results = int(results)
         except (ValueError, OverflowError):
             results = 20
-        if not isinstance(name, UnicodeType):
+        if not isinstance(name, unicode):
             name = unicode(name, encoding, 'replace')
         res = self._search_person(name, results)
         return [Person.Person(personID=self._get_real_personID(pi),
@@ -442,7 +484,7 @@ class IMDbBase:
             results = int(results)
         except (ValueError, OverflowError):
             results = 20
-        if not isinstance(name, UnicodeType):
+        if not isinstance(name, unicode):
             name = unicode(name, encoding, 'replace')
         res = self._search_character(name, results)
         return [Character.Character(characterID=self._get_real_characterID(pi),
@@ -487,22 +529,85 @@ class IMDbBase:
             results = int(results)
         except (ValueError, OverflowError):
             results = 20
-        if not isinstance(name, UnicodeType):
+        if not isinstance(name, unicode):
             name = unicode(name, encoding, 'replace')
         res = self._search_company(name, results)
         return [Company.Company(companyID=self._get_real_companyID(pi),
                 data=pd, modFunct=self._defModFunct,
                 accessSystem=self.accessSystem) for pi, pd in res][:results]
 
+    def _search_keyword(self, keyword, results):
+        """Return a list of 'keyword' strings."""
+        # XXX: for the real implementation, see the method of the
+        #      subclass, somewhere under the imdb.parser package.
+        raise NotImplementedError, 'override this method'
+
+    def search_keyword(self, keyword, results=None):
+        """Search for existing keywords, similar to the given one."""
+        if results is None:
+            results = self._keywordsResults
+        try:
+            results = int(results)
+        except (ValueError, OverflowError):
+            results = 100
+        if not isinstance(keyword, unicode):
+            keyword = unicode(keyword, encoding, 'replace')
+        return self._search_keyword(keyword, results)
+
+    def _get_keyword(self, keyword, results):
+        """Return a list of tuples (movieID, {movieData})"""
+        # XXX: for the real implementation, see the method of the
+        #      subclass, somewhere under the imdb.parser package.
+        raise NotImplementedError, 'override this method'
+
+    def get_keyword(self, keyword, results=None):
+        """Return a list of movies for the given keyword."""
+        if results is None:
+            results = self._keywordsResults
+        try:
+            results = int(results)
+        except (ValueError, OverflowError):
+            results = 100
+        # XXX: I suppose it will be much safer if the user provides
+        #      an unicode string... this is just a guess.
+        if not isinstance(keyword, unicode):
+            keyword = unicode(keyword, encoding, 'replace')
+        res = self._get_keyword(keyword, results)
+        return [Movie.Movie(movieID=self._get_real_movieID(mi),
+                data=md, modFunct=self._defModFunct,
+                accessSystem=self.accessSystem) for mi, md in res][:results]
+
+    def _get_top_bottom_movies(self, kind):
+        """Return the list of the top 250 or bottom 100 movies."""
+        # XXX: for the real implementation, see the method of the
+        #      subclass, somewhere under the imdb.parser package.
+        #      This method must return a list of (movieID, {movieDict})
+        #      tuples.  The kind parameter can be 'top' or 'bottom'.
+        raise NotImplementedError, 'override this method'
+
+    def get_top250_movies(self):
+        """Return the list of the top 250 movies."""
+        res = self._get_top_bottom_movies('top')
+        return [Movie.Movie(movieID=self._get_real_movieID(mi),
+                data=md, modFunct=self._defModFunct,
+                accessSystem=self.accessSystem) for mi, md in res]
+
+    def get_bottom100_movies(self):
+        """Return the list of the bottom 100 movies."""
+        res = self._get_top_bottom_movies('bottom')
+        return [Movie.Movie(movieID=self._get_real_movieID(mi),
+                data=md, modFunct=self._defModFunct,
+                accessSystem=self.accessSystem) for mi, md in res]
+
     def new_movie(self, *arguments, **keywords):
         """Return a Movie object."""
         # XXX: not really useful...
-        if keywords.has_key('title'):
-            if not isinstance(keywords['title'], UnicodeType):
+        if 'title' in keywords:
+            if not isinstance(keywords['title'], unicode):
                 keywords['title'] = unicode(keywords['title'],
                                             encoding, 'replace')
         elif len(arguments) > 1:
-            if not isinstance(arguments[1], UnicodeType):
+            if not isinstance(arguments[1], unicode):
                 arguments[1] = unicode(arguments[1], encoding, 'replace')
         return Movie.Movie(accessSystem=self.accessSystem,
                             *arguments, **keywords)
@@ -510,12 +615,12 @@ class IMDbBase:
     def new_person(self, *arguments, **keywords):
         """Return a Person object."""
         # XXX: not really useful...
-        if keywords.has_key('name'):
-            if not isinstance(keywords['name'], UnicodeType):
+        if 'name' in keywords:
+            if not isinstance(keywords['name'], unicode):
                 keywords['name'] = unicode(keywords['name'],
                                             encoding, 'replace')
         elif len(arguments) > 1:
-            if not isinstance(arguments[1], UnicodeType):
+            if not isinstance(arguments[1], unicode):
                 arguments[1] = unicode(arguments[1], encoding, 'replace')
         return Person.Person(accessSystem=self.accessSystem,
                                 *arguments, **keywords)
@@ -523,12 +628,12 @@ class IMDbBase:
     def new_character(self, *arguments, **keywords):
         """Return a Character object."""
         # XXX: not really useful...
-        if keywords.has_key('name'):
-            if not isinstance(keywords['name'], UnicodeType):
+        if 'name' in keywords:
+            if not isinstance(keywords['name'], unicode):
                 keywords['name'] = unicode(keywords['name'],
                                             encoding, 'replace')
         elif len(arguments) > 1:
-            if not isinstance(arguments[1], UnicodeType):
+            if not isinstance(arguments[1], unicode):
                 arguments[1] = unicode(arguments[1], encoding, 'replace')
         return Character.Character(accessSystem=self.accessSystem,
                                     *arguments, **keywords)
@@ -536,12 +641,12 @@ class IMDbBase:
     def new_company(self, *arguments, **keywords):
         """Return a Company object."""
         # XXX: not really useful...
-        if keywords.has_key('name'):
-            if not isinstance(keywords['name'], UnicodeType):
+        if 'name' in keywords:
+            if not isinstance(keywords['name'], unicode):
                 keywords['name'] = unicode(keywords['name'],
                                             encoding, 'replace')
         elif len(arguments) > 1:
-            if not isinstance(arguments[1], UnicodeType):
+            if not isinstance(arguments[1], unicode):
                 arguments[1] = unicode(arguments[1], encoding, 'replace')
         return Company.Company(accessSystem=self.accessSystem,
                                     *arguments, **keywords)
@@ -596,29 +701,45 @@ class IMDbBase:
                 info = self.get_character_infoset()
             else:
                 info = self.get_company_infoset()
-        if not isinstance(info, (TupleType, ListType)):
+        if not isinstance(info, (tuple, list)):
             info = (info,)
         res = {}
         for i in info:
-            if i in mop.current_info and not override: continue
+            if i in mop.current_info and not override:
+                continue
+            if not i:
+                continue
+            self._imdb_logger.debug('retrieving "%s" info set', i)
             try:
                 method = getattr(aSystem, 'get_%s_%s' %
                                     (prefix, i.replace(' ', '_')))
             except AttributeError:
-                raise IMDbDataAccessError, 'unknown information set "%s"' % i
-            ret = method(mopID)
-            if ret.has_key('info sets'):
-                for ri in ret['info sets']:
-                    mop.add_to_current_info(ri)
-            else:
-                mop.add_to_current_info(i)
-            if ret.has_key('data'):
+                self._imdb_logger.error('unknown information set "%s"', i)
+                # Keeps going.
+                method = lambda *x: {}
+            try:
+                ret = method(mopID)
+            except Exception, e:
+                self._imdb_logger.critical('caught an exception retrieving ' \
+                                    'or parsing "%s" info set for mopID ' \
+                                    '"%s" (accessSystem: %s)',
+                                    i, mopID, mop.accessSystem, exc_info=True)
+                ret = {}
+            keys = None
+            if 'data' in ret:
                 res.update(ret['data'])
-            if ret.has_key('titlesRefs'):
+                if isinstance(ret['data'], dict):
+                    keys = ret['data'].keys()
+            if 'info sets' in ret:
+                for ri in ret['info sets']:
+                    mop.add_to_current_info(ri, keys, mainInfoset=i)
+            else:
+                mop.add_to_current_info(i, keys)
+            if 'titlesRefs' in ret:
                 mop.update_titlesRefs(ret['titlesRefs'])
-            if ret.has_key('namesRefs'):
+            if 'namesRefs' in ret:
                 mop.update_namesRefs(ret['namesRefs'])
-            if ret.has_key('charactersRefs'):
+            if 'charactersRefs' in ret:
                 mop.update_charactersRefs(ret['charactersRefs'])
         mop.set_data(res, override=0)
 
@@ -650,133 +771,64 @@ class IMDbBase:
         #      subclass, somewhere under the imdb.parser package.
         raise NotImplementedError, 'override this method'
 
-    def _searchIMDb(self, params):
-        """Fetch the given search page from the IMDb akas server."""
-        from imdb.parser.http import IMDbURLopener
-        url = imdbURL_find % params
-        content = u''
+    def _searchIMDb(self, kind, ton):
+        """Search the IMDb akas server for the given title or name."""
+        # The Exact Primary search system has gone AWOL, so we resort
+        # to the mobile search. :-/
+        if not ton:
+            return None
+        aSystem = IMDb('mobile')
+        if kind == 'tt':
+            searchFunct = aSystem.search_movie
+            check = 'long imdb canonical title'
+        elif kind == 'nm':
+            searchFunct = aSystem.search_person
+            check = 'long imdb canonical name'
+        elif kind == 'char':
+            searchFunct = aSystem.search_character
+            check = 'long imdb canonical name'
+        elif kind == 'co':
+            # XXX: are [COUNTRY] codes included in the results?
+            searchFunct = aSystem.search_company
+            check = 'long imdb name'
         try:
-            urlOpener = IMDbURLopener()
-            content = urlOpener.retrieve_unicode(url)
-        except (IOError, IMDbDataAccessError):
-            pass
-        return content
+            searchRes = searchFunct(ton)
+        except IMDbError:
+            return None
+        # When only one result is returned, assume it was from an
+        # exact match.
+        if len(searchRes) == 1:
+            return searchRes[0].getID()
+        for item in searchRes:
+            # Return the first perfect match.
+            if item[check] == ton:
+                return item.getID()
+        return None
 
     def title2imdbID(self, title):
         """Translate a movie title (in the plain text data files format)
         to an imdbID.
         Try an Exact Primary Title search on IMDb;
         return None if it's unable to get the imdbID."""
-        if not title: return None
-        import urllib
-        if isinstance(title, unicode):
-            title = title.encode('utf-8')
-        params = 'q=%s;s=pt' % str(urllib.quote_plus(title))
-        content = self._searchIMDb(params)
-        if content and content[:512].find('<title>IMDb Title') != -1:
-            # Sometimes (e.g.: for titles with a "+" in it) a list
-            # of results is returned even for Exact Primary searches;
-            # this try to deal with it, hoping that a "normal" query
-            # will result in just on title.
-            params = 's=tt&q=%s' % str(urllib.quote_plus(title))
-            content = self._searchIMDb(params)
-        if not content: return None
-        from imdb.parser.http.searchMovieParser import DOMBasicMovieParser
-        mparser = DOMBasicMovieParser()
-        result = mparser.parse(content)
-        if not (result and result.get('data')): return None
-        return result['data'][0][0]
+        return self._searchIMDb('tt', title)
 
     def name2imdbID(self, name):
         """Translate a person name in an imdbID.
         Try an Exact Primary Name search on IMDb;
         return None if it's unable to get the imdbID."""
-        if not name: return None
-        import urllib
-        if isinstance(name, unicode):
-            name = name.encode('utf-8')
-        params = 'q=%s;s=pn' % str(urllib.quote_plus(name))
-        content = self._searchIMDb(params)
-        if content and content[:512].find('<title>IMDb Name') != -1:
-            params = 's=nm&q=%s' % str(urllib.quote_plus(name))
-            content = self._searchIMDb(params)
-        if not content: return None
-        from imdb.parser.http.searchPersonParser import DOMBasicPersonParser
-        pparser = DOMBasicPersonParser()
-        result = pparser.parse(content)
-        if not (result and result.get('data')): return None
-        return result['data'][0][0]
+        return self._searchIMDb('tt', name)
 
     def character2imdbID(self, name):
         """Translate a character name in an imdbID.
         Try an Exact Primary Name search on IMDb;
         return None if it's unable to get the imdbID."""
-        if not name: return None
-        import urllib
-        if isinstance(name, unicode):
-            name = name.encode('utf-8')
-        # XXX: s=pc is just a assumption of mine; it doesn't seem
-        #      to work, and I can't find a working alternative.
-        params = 'q=%s;s=pc' % str(urllib.quote_plus(name))
-        content = self._searchIMDb(params)
-        if not content: return None
-        if content[:512].find('<title>IMDb  Search') != -1:
-            from imdb.parser.http.searchCharacterParser import \
-                        DOMHTMLSearchCharacterParser
-            search_character_parser = DOMHTMLSearchCharacterParser()
-            result = search_character_parser.parse(content)
-            if not result: return None
-            if not result.has_key('data'): return None
-            if not result['data']: return None
-            # Try to read the first result.
-            chID, dname = result['data'][0]
-            if isinstance(dname.get('name'), unicode):
-                rname = dname['name'].encode('utf-8')
-            if name == rname:
-                return chID
-            return None
-        # XXX: still needed?
-        from imdb.parser.http.searchCharacterParser import \
-                    DOMBasicCharacterParser
-        cparser = DOMBasicCharacterParser()
-        result = cparser.parse(content)
-        if not (result and result.get('data')): return None
-        return result['data'][0][0]
+        return self._searchIMDb('char', name)
 
     def company2imdbID(self, name):
         """Translate a company name in an imdbID.
         Try an Exact Primary Name search on IMDb;
         return None if it's unable to get the imdbID."""
-        if not name: return None
-        import urllib
-        if isinstance(name, unicode):
-            name = name.encode('utf-8')
-        # XXX: s=pco is just a assumption of mine; it doesn't seem
-        #      to work, and I can't find a working alternative.
-        params = 'q=%s;s=pco' % str(urllib.quote_plus(name))
-        content = self._searchIMDb(params)
-        if not content: return None
-        if content[:512].find('<title>IMDb  Search') != -1:
-            from imdb.parser.http.searchCompanyParser import \
-                        DOMHTMLSearchCompanyParser
-            search_company_parser = DOMHTMLSearchCompanyParser()
-            result = search_company_parser.parse(content)
-            if not result: return None
-            if not result.has_key('data'): return None
-            if not result['data']: return None
-            # Try to read the first result.
-            chID, dname = result['data'][0]
-            if isinstance(dname.get('name'), unicode):
-                rname = dname['name'].encode('utf-8')
-            if name == rname:
-                return chID
-            return None
-        # Still needed?
-        from imdb.parser.http.searchCompanyParser import DOMBasicCompanyParser
-        cparser = DOMBasicCompanyParser()
-        result = cparser.parse(content)
-        if not (result and result.get('data')): return None
-        return result['data'][0][0]
+        return self._searchIMDb('co', name)
 
     def get_imdbID(self, mop):
         """Return the imdbID for the given Movie, Person, Character or Company
@@ -790,7 +842,7 @@ class IMDbBase:
             if mop.movieID is not None:
                 imdbID = aSystem.get_imdbMovieID(mop.movieID)
             else:
-                imdbID = aSystem.title2imdbID(build_title(mop, canonical=1,
+                imdbID = aSystem.title2imdbID(build_title(mop, canonical=0,
                                                 ptdf=1))
         elif isinstance(mop, Person.Person):
             if mop.personID is not None:
@@ -801,6 +853,7 @@ class IMDbBase:
             if mop.characterID is not None:
                 imdbID = aSystem.get_imdbCharacterID(mop.characterID)
             else:
+                # canonical=0 ?
                 imdbID = aSystem.character2imdbID(build_name(mop, canonical=1))
         elif isinstance(mop, Company.Company):
             if mop.companyID is not None:
