@@ -4,7 +4,7 @@ parser.http.utils module (imdb package).
 This module provides miscellaneous utilities used by
 the imdb.parser.http classes.
 
-Copyright 2004-2008 Davide Alberani <da@erlug.linux.it>
+Copyright 2004-2010 Davide Alberani <da@erlug.linux.it>
                2008 H. Turgut Uyar <uyar@tekir.org>
 
 This program is free software; you can redistribute it and/or modify
@@ -23,12 +23,9 @@ Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 """
 
 import re
-import warnings
-from types import UnicodeType, StringType, ListType, DictType
-from sgmllib import SGMLParser
-from urllib import unquote
+import logging
 
-from imdb._exceptions import IMDbParserError, IMDbError
+from imdb._exceptions import IMDbError
 
 from imdb.utils import flatten, _Container
 from imdb.Movie import Movie
@@ -56,9 +53,9 @@ _modify_keys = list(Movie.keys_tomodify_list) + list(Person.keys_tomodify_list)
 def _putRefs(d, re_titles, re_names, re_characters, lastKey=None):
     """Iterate over the strings inside list items or dictionary values,
     substitutes movie titles and person names with the (qv) references."""
-    if isinstance(d, ListType):
+    if isinstance(d, list):
         for i in xrange(len(d)):
-            if isinstance(d[i], (UnicodeType, StringType)):
+            if isinstance(d[i], (unicode, str)):
                 if lastKey in _modify_keys:
                     if re_names:
                         d[i] = re_names.sub(ur"'\1' (qv)", d[i])
@@ -66,13 +63,13 @@ def _putRefs(d, re_titles, re_names, re_characters, lastKey=None):
                         d[i] = re_titles.sub(ur'_\1_ (qv)', d[i])
                     if re_characters:
                         d[i] = re_characters.sub(ur'#\1# (qv)', d[i])
-            elif isinstance(d[i], (ListType, DictType)):
+            elif isinstance(d[i], (list, dict)):
                 _putRefs(d[i], re_titles, re_names, re_characters,
                         lastKey=lastKey)
-    elif isinstance(d, DictType):
+    elif isinstance(d, dict):
         for k, v in d.items():
             lastKey = k
-            if isinstance(v, (UnicodeType, StringType)):
+            if isinstance(v, (unicode, str)):
                 if lastKey in _modify_keys:
                     if re_names:
                         d[k] = re_names.sub(ur"'\1' (qv)", v)
@@ -80,7 +77,7 @@ def _putRefs(d, re_titles, re_names, re_characters, lastKey=None):
                         d[k] = re_titles.sub(ur'_\1_ (qv)', v)
                     if re_characters:
                         d[k] = re_characters.sub(ur'#\1# (qv)', v)
-            elif isinstance(v, (ListType, DictType)):
+            elif isinstance(v, (list, dict)):
                 _putRefs(d[k], re_titles, re_names, re_characters,
                         lastKey=lastKey)
 
@@ -91,7 +88,7 @@ entitydefs = entitydefs.copy()
 entitydefsget = entitydefs.get
 entitydefs['nbsp'] = ' '
 
-sgmlentity = SGMLParser.entitydefs.copy()
+sgmlentity = {'lt': '<', 'gt': '>', 'amp': '&', 'quot': '"', 'apos': '\''}
 sgmlentityget = sgmlentity.get
 _sgmlentkeys = sgmlentity.keys()
 
@@ -110,9 +107,20 @@ for _k, _v in entitydefs.items():
     entcharrefs[_k] = _v
 del _sgmlentkeys, _k, _v
 entcharrefs['#160'] = u' '
+entcharrefs['#xA0'] = u' '
+entcharrefs['#xa0'] = u' '
+entcharrefs['#XA0'] = u' '
+entcharrefs['#x22'] = u'"'
+entcharrefs['#X22'] = u'"'
+# convert &x26; to &amp;, to make BeautifulSoup happy; beware that this
+# leaves lone '&' in the html broken, but I assume this is better than
+# the contrary...
+entcharrefs['#38'] = u'&amp;'
+entcharrefs['#x26'] = u'&amp;'
+entcharrefs['#x26'] = u'&amp;'
 
-re_entcharrefs = re.compile('&(%s|\#160|\#\d{1,5});' %
-                            '|'.join(map(re.escape, entcharrefs)))
+re_entcharrefs = re.compile('&(%s|\#160|\#\d{1,5}|\#x[0-9a-f]{1,4});' %
+                            '|'.join(map(re.escape, entcharrefs)), re.I)
 re_entcharrefssub = re_entcharrefs.sub
 
 sgmlentity.update(dict([('#34', u'"'), ('#38', u'&'),
@@ -133,8 +141,15 @@ def _replXMLRef(match):
     if value is None:
         if ref[0] == '#':
             ref_code = ref[1:]
-            if ref_code in ('34', '38', '60', '62', '39'): return match.group(0)
-            else: return unichr(int(ref[1:]))
+            if ref_code in ('34', '38', '60', '62', '39'):
+                return match.group(0)
+            elif ref_code[0].lower() == 'x':
+                #if ref[2:] == '26':
+                #    # Don't convert &x26; to &amp;, to make BeautifulSoup happy.
+                #    return '&amp;'
+                return unichr(int(ref[2:], 16))
+            else:
+                return unichr(int(ref[1:]))
         else:
             return ref
     return value
@@ -144,6 +159,7 @@ def subXMLRefs(s):
     replaced."""
     return re_entcharrefssub(_replXMLRef, s)
 
+# XXX: no more used here; move it to mobile (they are imported by helpers, too)?
 def _replSGMLRefs(match):
     """Replace the matched SGML entity."""
     ref = match.group(1)
@@ -155,10 +171,13 @@ def subSGMLRefs(s):
     return re_sgmlrefsub(_replSGMLRefs, s)
 
 
+_b_p_logger = logging.getLogger('imdbpy.parser.http.build_person')
 def build_person(txt, personID=None, billingPos=None,
                 roleID=None, accessSystem='http', modFunct=None):
     """Return a Person instance from the tipical <tr>...</tr> strings
     found in the IMDb's web site."""
+    #if personID is None
+    #    _b_p_logger.debug('empty name or personID for "%s"', txt)
     notes = u''
     role = u''
     # Search the (optional) separator between name and role/notes.
@@ -184,17 +203,35 @@ def build_person(txt, personID=None, billingPos=None,
         elif role_comment[-6:] == '& ....':
             role_comment = role_comment[:-6].rstrip()
         # Get the notes.
-        cmt_idx = role_comment.find('(')
-        if cmt_idx != -1:
-            role = role_comment[:cmt_idx].rstrip()
-            notes = role_comment[cmt_idx:]
+        if roleID is not None:
+            if not isinstance(roleID, list):
+                cmt_idx = role_comment.find('(')
+                if cmt_idx != -1:
+                    role = role_comment[:cmt_idx].rstrip()
+                    notes = role_comment[cmt_idx:]
+                else:
+                    # Just a role, without notes.
+                    role = role_comment
+            else:
+                role = role_comment
         else:
-            # Just a role, without notes.
-            role = role_comment
+            # We're managing something that doesn't have a 'role', so
+            # everything are notes.
+            notes = role_comment
     if role == '....': role = u''
+    roleNotes = []
     # Manages multiple roleIDs.
     if isinstance(roleID, list):
-        role = role.split(' / ')
+        rolesplit = role.split('/')
+        role = []
+        for r in rolesplit:
+            nidx = r.find('(')
+            if nidx != -1:
+                role.append(r[:nidx].rstrip())
+                roleNotes.append(r[nidx:])
+            else:
+                role.append(r)
+                roleNotes.append(None)
         lr = len(role)
         lrid = len(roleID)
         if lr > lrid:
@@ -211,44 +248,57 @@ def build_person(txt, personID=None, billingPos=None,
         roleID = str(roleID)
     if personID is not None:
         personID = str(personID)
+    if (not name) or (personID is None):
+        # Set to 'debug', since build_person is expected to receive some crap.
+        _b_p_logger.debug('empty name or personID for "%s"', txt)
     # XXX: return None if something strange is detected?
-    return Person(name=name, personID=personID, currentRole=role,
+    person = Person(name=name, personID=personID, currentRole=role,
                     roleID=roleID, notes=notes, billingPos=billingPos,
                     modFunct=modFunct, accessSystem=accessSystem)
+    if roleNotes and len(roleNotes) == len(roleID):
+        for idx, role in enumerate(person.currentRole):
+            if roleNotes[idx]:
+                role.notes = roleNotes[idx]
+    return person
 
 
+_re_chrIDs = re.compile('[0-9]{7}')
+
+_b_m_logger = logging.getLogger('imdbpy.parser.http.build_movie')
 # To shrink spaces.
 re_spaces = re.compile(r'\s+')
 def build_movie(txt, movieID=None, roleID=None, status=None,
                 accessSystem='http', modFunct=None, _parsingCharacter=False,
-                _parsingCompany=False):
+                _parsingCompany=False, year=None, chrRoles=None,
+                rolesNoChar=None, additionalNotes=None):
     """Given a string as normally seen on the "categorized" page of
     a person on the IMDb's web site, returns a Movie instance."""
+    # FIXME: Oook, lets face it: build_movie and build_person are now
+    # two horrible sets of patches to support the new IMDb design.  They
+    # must be rewritten from scratch.
     if _parsingCharacter:
         _defSep = ' Played by '
     elif _parsingCompany:
         _defSep = ' ... '
     else:
-        _defSep = '....'
+        _defSep = ' .... '
     title = re_spaces.sub(' ', txt).strip()
     # Split the role/notes from the movie title.
     tsplit = title.split(_defSep, 1)
     role = u''
     notes = u''
+    roleNotes = []
     if len(tsplit) == 2:
         title = tsplit[0].rstrip()
         role = tsplit[1].lstrip()
-        # Find notes in the role.
-        if role[-1:] == ')':
-            nidx = role.find('(')
-            # XXX: check balanced parentheses?
-            if nidx != -1:
-                notes = role[nidx:]
-                role = role[:nidx].rstrip()
     if title[-9:] == 'TV Series':
         title = title[:-9].rstrip()
+    elif title[-14:] == 'TV mini-series':
+        title = title[:-14] + ' (mini)'
     # Try to understand where the movie title ends.
     while True:
+        if year:
+            break
         if title[-1:] != ')':
             # Ignore the silly "TV Series" notice.
             if title[-9:] == 'TV Series':
@@ -277,15 +327,36 @@ def build_movie(txt, movieID=None, roleID=None, status=None,
         if notes: notes = '%s %s' % (title[nidx:], notes)
         else: notes = title[nidx:]
         title = title[:nidx].rstrip()
+    if year:
+        year = year.strip()
+        if title[-1] == ')':
+            fpIdx = title.rfind('(')
+            if fpIdx != -1:
+                if notes: notes = '%s %s' % (title[fpIdx:], notes)
+                else: notes = title[fpIdx:]
+                title = title[:fpIdx].rstrip()
+        title = u'%s (%s)' % (title, year)
     if _parsingCharacter and roleID and not role:
         roleID = None
     if not roleID:
         roleID = None
     elif len(roleID) == 1:
         roleID = roleID[0]
+    if not role and chrRoles and isinstance(roleID, (str, unicode)):
+        roleID = _re_chrIDs.findall(roleID)
+        role = ' / '.join(filter(None, chrRoles.split('@@')))
     # Manages multiple roleIDs.
     if isinstance(roleID, list):
-        role = role.split(' / ')
+        tmprole = role.split('/')
+        role = []
+        for r in tmprole:
+            nidx = r.find('(')
+            if nidx != -1:
+                role.append(r[:nidx].rstrip())
+                roleNotes.append(r[nidx:])
+            else:
+                role.append(r)
+                roleNotes.append(None)
         lr = len(role)
         lrid = len(roleID)
         if lr > lrid:
@@ -302,307 +373,53 @@ def build_movie(txt, movieID=None, roleID=None, status=None,
         roleID = str(roleID)
     if movieID is not None:
         movieID = str(movieID)
+    if (not title) or (movieID is None):
+        _b_m_logger.error('empty title or movieID for "%s"', txt)
+    if rolesNoChar:
+        rolesNoChar = filter(None, [x.strip() for x in rolesNoChar.split('/')])
+        if not role:
+            role = []
+        elif not isinstance(role, list):
+            role = [role]
+        role += rolesNoChar
+    notes = notes.strip()
+    if additionalNotes:
+        additionalNotes = re_spaces.sub(' ', additionalNotes).strip()
+        if notes:
+            notes += u' '
+        notes += additionalNotes
     m = Movie(title=title, movieID=movieID, notes=notes, currentRole=role,
                 roleID=roleID, roleIsPerson=_parsingCharacter,
                 modFunct=modFunct, accessSystem=accessSystem)
+    if roleNotes and len(roleNotes) == len(roleID):
+        for idx, role in enumerate(m.currentRole):
+            try:
+                if roleNotes[idx]:
+                    role.notes = roleNotes[idx]
+            except IndexError:
+                break
     # Status can't be checked here, and must be detected by the parser.
     if status:
         m['status'] = status
     return m
 
 
-# XXX: this class inherits from SGMLParser; see the documentation for
-#      the "sgmllib" modules.
-class ParserBase(SGMLParser):
-    """Base parser to handle HTML data from the IMDb's web server."""
-    # The imdbID is a 7-ciphers number.
-    re_imdbID = re.compile(r'(?<=nm|tt|ch|co)([0-9]{7})\b')
-    re_imdbIDonly = re.compile(r'\b([0-9]{7})\b')
-    re_airdate = re.compile(r'(.*)\s*\(season (\d+), episode (\d+)\)', re.I)
-    _re_imdbIDmatch = re.compile(r'(nm|tt|ch|co)[0-9]{7}/?$')
-
-    # It's set when names and titles references must be collected.
-    # It can be set to 0 for search parsers.
-    _defGetRefs = False
-    entitydefs = sgmlentity
-
-    def __init__(self, verbose=0):
-        self._init()
-        # Fall-back defaults.
-        self._modFunct = None
-        self._as = 'http'
-        SGMLParser.__init__(self, verbose)
-
-    def handle_charref(self, name):
-        # Handles "quotes", "less than", "greater than" and so on.
-        try:
-            ret = unichr(int(name))
-            self.handle_data(ret)
-            return
-        except (ValueError, TypeError, OverflowError):
-            pass
-        return SGMLParser.handle_charref(self, name)
-
-    def unknown_charref(self, ref):
-        try:
-            n = unichr(int(ref))
-            self.handle_data(n)
-        except (TypeError, ValueError, OverflowError):
-            return SGMLParser.unknown_charref(self, ref)
-
-    def reset(self):
-        """Reset the parser."""
-        SGMLParser.reset(self)
-        # Names and titles references.
-        self._namesRefs = {}
-        self._titlesRefs = {}
-        self._charactersRefs = {}
-        self._titleRefCID = u''
-        self._nameRefCID = u''
-        self._characterRefCID = u''
-        self._titleCN = u''
-        self._nameCN = u''
-        self._characterCN = u''
-        self._inTTRef = 0
-        self._inLinkTTRef = 0
-        self._inNMRef = 0
-        self._inCHRef = 0
-        self._in_content = 0
-        self._div_count = 0
-        self._reset()
-
-    def get_attr_value(self, attrs_list, searched_attr):
-        """Given a list of attributes in the form ('attr_name', 'attr_value')',
-        return the attr_value of the 'searched_attr' attribute or None if it's
-        not found."""
-        for attr in attrs_list:
-            if attr[0] == searched_attr:
-                attr = attr[1]
-                try:
-                    attr = unquote(str(attr))
-                    attr = unicode(attr, 'latin_1')
-                except UnicodeEncodeError:
-                    pass
-                return subSGMLRefs(attr)
-        return None
-
-    def _init(self): pass
-
-    def _reset(self): pass
-
-    def get_data(self): return None
-
-    def handle_data(self, data):
-        """Gather information about movie titles and person names,
-        and call the _handle_data method."""
-        if self.getRefs:
-            if self._inNMRef:
-                self._nameCN += data.replace('\n', ' ')
-            elif self._inTTRef:
-                if self._inLinkTTRef:
-                    self._titleCN += data.replace('\n', ' ')
-                else:
-                    sdata = data.strip().replace('\n', ' ')
-                    yearK = re_yearKind_index.match(sdata)
-                    if yearK and yearK.start() == 0:
-                        self._titleCN += ' %s' % sdata[:yearK.end()]
-                        self._add_ref('tt')
-            elif self._inCHRef:
-                self._characterCN += data.replace('\n', ' ')
-        self._handle_data(data)
-
-    def _handle_data(self, data): pass
-
-    def _add_ref(self, kind):
-        """Add a reference entry to the names and titles dictionaries."""
-        if kind == 'tt':
-            if self._titleRefCID and self._titleCN:
-                if not self._titlesRefs.has_key(self._titleCN):
-                    try:
-                        movie = Movie(movieID=str(self._titleRefCID),
-                                    title=self._titleCN, accessSystem=self._as,
-                                    modFunct=self._modFunct)
-                        self._titlesRefs[self._titleCN] = movie
-                    except IMDbParserError:
-                        pass
-                self._titleRefCID = u''
-                self._titleCN = u''
-                self._inTTRef = 0
-                self._inLinkTTRef = 0
-        elif kind == 'nm' and self._nameRefCID and self._nameCN:
-            # XXX: 'Neo' and 'Keanu Reeves' are two separated
-            #      entry in the dictionary.  Check the ID value instead
-            #      of the key?
-            if not self._namesRefs.has_key(self._nameCN):
-                try:
-                    person = Person(name=self._nameCN,
-                                    personID=str(self._nameRefCID),
-                                    accessSystem=self._as,
-                                    modFunct=self._modFunct)
-                    self._namesRefs[self._nameCN] = person
-                except IMDbParserError:
-                    pass
-            self._nameRefCID = u''
-            self._nameCN = u''
-            self._inNMRef = 0
-        elif kind == 'ch' and self._characterRefCID and self._characterCN:
-            if not self._charactersRefs.has_key(self._characterCN):
-                try:
-                    character = Character(name=self._characterCN,
-                                    characterID=str(self._characterRefCID),
-                                    accessSystem='http')
-                    self._charactersRefs[self._characterCN] = character
-                except IMDbParserError:
-                    pass
-            self._characterRefCID = u''
-            self._characterCN = u''
-            self._inCHRef = 0
-
-    def _refs_anchor_bgn(self, attrs):
-        """At the start of an 'a' tag, gather info for the
-        references dictionaries."""
-        if self._inTTRef: self._add_ref('tt')
-        if self._inNMRef: self._add_ref('nm')
-        if self._inCHRef: self._add_ref('ch')
-        href = self.get_attr_value(attrs, 'href')
-        if not href: return
-        if href.startswith('/title/tt'):
-            href = href[7:]
-            if not self._re_imdbIDmatch.match(href) or \
-                    (len(href) > 10 and href[10:11] != '?'):
-                return
-            href = href[2:]
-            if href[-1] == '/': href = href[:-1]
-            self._titleRefCID = href
-            self._inTTRef = 1
-            self._inLinkTTRef = 1
-        elif href.startswith('/name/nm'):
-            href = href[6:]
-            if not self._re_imdbIDmatch.match(href) or \
-                    (len(href) > 10 and href[10:11] != '?'):
-                return
-            href = href[2:]
-            if href[-1] == '/': href = href[:-1]
-            self._nameRefCID = href
-            self._inNMRef = 1
-        elif href.startswith('/character/ch'):
-            href = href[11:]
-            if not self._re_imdbIDmatch.match(href):
-                return
-            href = href[2:]
-            if href[-1] == '/': href = href[:-1]
-            self._characterRefCID = href
-            self._inCHRef = 1
-
-    def _refs_anchor_end(self):
-        """At the end of an 'a' tag, gather info for the
-        references dictionaries."""
-        # XXX: check if self.getRefs is True?
-        self._add_ref('nm')
-        self._add_ref('ch')
-        self._inLinkTTRef = 0
-
-    def handle_starttag(self, tag, method, attrs):
-        if self.getRefs:
-            # XXX: restrict collection to links in self._in_content ?
-            if tag == 'a': self._refs_anchor_bgn(attrs)
-        if tag == 'div':
-            if not self._in_content:
-                # In the new IMDb's layout the content is nicely tagged. :-)
-                if self.get_attr_value(attrs, 'id') == 'tn15content':
-                    self._in_content = 1
-                    self._div_count = 1
-                    self._begin_content()
-            else:
-                # Another div tag inside the content.
-                self._div_count += 1
-        method(attrs)
-
-    def handle_endtag(self, tag, method):
-        if self.getRefs:
-            if tag == 'a': self._refs_anchor_end()
-        # Count div tags inside the 'tn15content' one, and set
-        # self._in_content to False only when the count drops to zero.
-        if tag == 'div':
-            if self._in_content:
-                self._div_count -= 1
-                if self._div_count <= 0:
-                    self._end_content()
-                    self._in_content = 0
-        method()
-
-    def _begin_content(self): pass
-    def _end_content(self): pass
-
-    def start_a(self, attrs): pass
-    def end_a(self): pass
-
-    def start_div(self, attrs): pass
-    def end_div(self): pass
-
-    def anchor_bgn(self, href, name, type): pass
-    def anchor_end(self): pass
-
-    def handle_image(self, src, alt, *args): pass
-
-    def error(self, message):
-        raise IMDbParserError, 'HTML parser error: "%s"' % str(message)
-
-    def parse(self, html_string, getRefs=None, **kwds):
-        """Return the dictionary generated from the given html string."""
-        self.reset()
-        if getRefs is not None:
-            self.getRefs = getRefs
-        else:
-            self.getRefs = self._defGetRefs
-        for key, value in kwds.items():
-            setattr(self, key, value)
-        # XXX: useful only for the testsuite.
-        if not isinstance(html_string, UnicodeType):
-            html_string = unicode(html_string, 'latin_1', 'replace')
-        html_string = subXMLRefs(html_string)
-        # Fix invalid HTML single tags like <br/>
-        html_string = re_xmltags.sub('<\\1 />', html_string)
-        self.feed(html_string)
-        # Fallback measure for wrong HTML - not sure why, but here
-        # self._in_content seems to be always False.
-        if self._div_count > 0:
-            self._end_content()
-            self._in_content = 0
-        if self.getRefs and self._inTTRef: self._add_ref('tt')
-        data = self.get_data()
-        if self.getRefs:
-            # It would be nice to use ur'\b(%s)\b', but it seems that
-            # some references are lost.
-            titl_re = ur'(%s)' % '|'.join([re.escape(x) for x
-                                            in self._titlesRefs.keys()])
-            if titl_re != ur'()': re_titles = re.compile(titl_re, re.U)
-            else: re_titles = None
-            nam_re = ur'(%s)' % '|'.join([re.escape(x) for x
-                                            in self._namesRefs.keys()])
-            if nam_re != ur'()': re_names = re.compile(nam_re, re.U)
-            else: re_names = None
-            chr_re = ur'(%s)' % '|'.join([re.escape(x) for x
-                                            in self._charactersRefs.keys()])
-            if chr_re != ur'()': re_characters = re.compile(chr_re, re.U)
-            else: re_characters = None
-            _putRefs(data, re_titles, re_names, re_characters)
-        # XXX: should I return a copy of data?  Answer: NO!
-        return {'data': data, 'titlesRefs': self._titlesRefs,
-                'namesRefs': self._namesRefs,
-                'charactersRefs': self._charactersRefs}
-
-
 class DOMParserBase(object):
-    """Base parser to handle HTML data from the IMDb's web server -
-    DOM/XPath version."""
+    """Base parser to handle HTML data from the IMDb's web server."""
     _defGetRefs = False
     _containsObjects = False
 
     preprocessors = []
     extractors = []
+    usingModule = None
+
+    _logger = logging.getLogger('imdbpy.parser.http.domparser')
 
     def __init__(self, useModule=None):
+        """Initialize the parser. useModule can be used to force it
+        to use 'BeautifulSoup' or 'lxml'; by default, it's auto-detected,
+        using 'lxml' if available and falling back to 'BeautifulSoup'
+        otherwise."""
         # Module to use.
         if useModule is None:
             useModule = ('lxml', 'BeautifulSoup')
@@ -615,42 +432,31 @@ class DOMParserBase(object):
             mod = mod.strip().lower()
             try:
                 if mod == 'lxml':
-                    from lxmladapter import fromstring
-                    from lxmladapter import tostring
-                    from lxmladapter import setattribute
-                    from lxmladapter import getattribute
-                    from lxmladapter import getparent
-                    from lxmladapter import clone
-                    from lxmladapter import apply_xpath
+                    from lxml.html import fromstring
+                    from lxml.etree import tostring
+                    self._is_xml_unicode = False
+                    self.usingModule = 'lxml'
                 elif mod == 'beautifulsoup':
-                    from bsoupadapter import fromstring
-                    from bsoupadapter import tostring
-                    from bsoupadapter import setattribute
-                    from bsoupadapter import getattribute
-                    from bsoupadapter import getparent
-                    from bsoupadapter import clone
-                    from bsoupadapter import apply_xpath
+                    from bsouplxml.html import fromstring
+                    from bsouplxml.etree import tostring
+                    self._is_xml_unicode = True
+                    self.usingModule = 'beautifulsoup'
                 else:
-                    warnings.warn('unknown module "%s".' % mod)
+                    self._logger.warn('unknown module "%s"' % mod)
                     continue
                 self.fromstring = fromstring
-                self.tostring = tostring
-                self.setattribute = setattribute
-                self.getattribute = getattribute
-                self.getparent = getparent
-                self.clone = clone
-                self.apply_xpath = apply_xpath
+                self._tostring = tostring
                 if _gotError:
-                    warnings.warn('falling back to "%s".' % mod)
+                    self._logger.warn('falling back to "%s"' % mod)
                 break
             except ImportError, e:
                 if idx+1 >= nrMods:
                     # Raise the exception, if we don't have any more
                     # options to try.
-                    raise IMDbError, 'unable to use any parser in %s: %s' % (
-                                                    str(useModule), str(e))
+                    raise IMDbError, 'unable to use any parser in %s: %s' % \
+                                    (str(useModule), str(e))
                 else:
-                    warnings.warn('unable to use "%s": %s' % (mod, str(e)))
+                    self._logger.warn('unable to use "%s": %s' % (mod, str(e)))
                     _gotError = True
                 continue
         else:
@@ -658,6 +464,7 @@ class DOMParserBase(object):
         # Fall-back defaults.
         self._modFunct = None
         self._as = 'http'
+        self._cname = self.__class__.__name__
         self._init()
         self.reset()
 
@@ -670,51 +477,117 @@ class DOMParserBase(object):
         self._reset()
 
     def _init(self):
+        """Subclasses can override this method, if needed."""
         pass
 
     def _reset(self):
+        """Subclasses can override this method, if needed."""
         pass
 
     def parse(self, html_string, getRefs=None, **kwds):
-        """Return the dictionary generated from the given html string."""
+        """Return the dictionary generated from the given html string;
+        getRefs can be used to force the gathering of movies/persons/characters
+        references."""
         self.reset()
         if getRefs is not None:
             self.getRefs = getRefs
         else:
             self.getRefs = self._defGetRefs
-        # XXX: useful only for the testsuite.
-        if not isinstance(html_string, UnicodeType):
+        # Useful only for the testsuite.
+        if not isinstance(html_string, unicode):
             html_string = unicode(html_string, 'latin_1', 'replace')
         html_string = subXMLRefs(html_string)
-        ## Not required?
-        ##html_string = subSGMLRefs(html_string)
         # Temporary fix: self.parse_dom must work even for empty strings.
         html_string = self.preprocess_string(html_string)
         html_string = html_string.strip()
+        # tag attributes like title="&#x22;Family Guy&#x22;" will be
+        # converted to title=""Family Guy"" and this confuses BeautifulSoup.
+        if self.usingModule == 'beautifulsoup':
+            html_string = html_string.replace('""', '"')
+        #print html_string.encode('utf8')
         if html_string:
             dom = self.get_dom(html_string)
-            dom = self.preprocess_dom(dom)
+            #print self.tostring(dom).encode('utf8')
+            try:
+                dom = self.preprocess_dom(dom)
+            except Exception, e:
+                self._logger.error('%s: caught exception preprocessing DOM',
+                                    self._cname, exc_info=True)
             if self.getRefs:
-                self.gather_refs(dom)
+                try:
+                    self.gather_refs(dom)
+                except Exception, e:
+                    self._logger.warn('%s: unable to gather refs: %s',
+                                    self._cname, exc_info=True)
             data = self.parse_dom(dom)
         else:
             data = {}
-        data = self.postprocess_data(data)
+        try:
+            data = self.postprocess_data(data)
+        except Exception, e:
+            self._logger.error('%s: caught exception postprocessing data',
+                                self._cname, exc_info=True)
         if self._containsObjects:
             self.set_objects_params(data)
         data = self.add_refs(data)
         return data
 
-    def get_dom(self, html_string):
-        return self.fromstring(html_string)
+    def _build_empty_dom(self):
+        from bsouplxml import _bsoup
+        return _bsoup.BeautifulSoup('')
 
-    def xpath(self, element, xpath):
-        return self.apply_xpath(element, xpath)
+    def get_dom(self, html_string):
+        """Return a dom object, from the given string."""
+        try:
+            dom = self.fromstring(html_string)
+            if dom is None:
+                dom = self._build_empty_dom()
+                self._logger.error('%s: using a fake empty DOM', self._cname)
+            return dom
+        except Exception, e:
+            self._logger.error('%s: caught exception parsing DOM',
+                                self._cname, exc_info=True)
+            return self._build_empty_dom()
+
+    def xpath(self, element, path):
+        """Return elements matching the given XPath."""
+        try:
+            xpath_result = element.xpath(path)
+            if self._is_xml_unicode:
+                return xpath_result
+            result = []
+            for item in xpath_result:
+                if isinstance(item, str):
+                    item = unicode(item)
+                result.append(item)
+            return result
+        except Exception, e:
+            self._logger.error('%s: caught exception extracting XPath "%s"',
+                                self._cname, path, exc_info=True)
+            return []
+
+    def tostring(self, element):
+        """Convert the element to a string."""
+        if isinstance(element, (unicode, str)):
+            return unicode(element)
+        else:
+            try:
+                return self._tostring(element, encoding=unicode)
+            except Exception, e:
+                self._logger.error('%s: unable to convert to string',
+                                    self._cname, exc_info=True)
+                return u''
+
+    def clone(self, element):
+        """Clone an element."""
+        return self.fromstring(self.tostring(element))
 
     def preprocess_string(self, html_string):
         """Here we can modify the text, before it's parsed."""
-        if len(html_string) == 0:
+        if not html_string:
             return html_string
+        # Remove silly &nbsp;&raquo; chars.
+        html_string = html_string.replace(u' \xbb', u'')
         try:
             preprocessors = self.preprocessors
         except AttributeError:
@@ -726,7 +599,12 @@ class DOMParserBase(object):
             elif isinstance(src, str):
                 html_string = html_string.replace(src, sub)
             elif callable(src):
-                html_string = src(html_string)
+                try:
+                    html_string = src(html_string)
+                except Exception, e:
+                    _msg = '%s: caught exception preprocessing html'
+                    self._logger.error(_msg, self._cname, exc_info=True)
+                    continue
         ##print html_string.encode('utf8')
         return html_string
 
@@ -751,6 +629,7 @@ class DOMParserBase(object):
         in self.extractors."""
         result = {}
         for extractor in self.extractors:
+            ##print extractor.label
             if extractor.group is None:
                 elements = [(extractor.label, element)
                             for element in self.xpath(dom, extractor.path)]
@@ -768,7 +647,12 @@ class DOMParserBase(object):
                     normalizer = extractor.group_key_normalize
                     if normalizer is not None:
                         if callable(normalizer):
-                            group_key = normalizer(group_key)
+                            try:
+                                group_key = normalizer(group_key)
+                            except Exception, e:
+                                _m = '%s: unable to apply group_key normalizer'
+                                self._logger.error(_m, self._cname,
+                                                    exc_info=True)
                     group_elements = self.xpath(group, extractor.path)
                     elements.extend([(group_key, element)
                                      for element in group_elements])
@@ -782,6 +666,7 @@ class DOMParserBase(object):
                             if not value:
                                 data[field] = None
                             else:
+                                # XXX: use u'' , to join?
                                 data[field] = ''.join(value)
                     else:
                         data = self.xpath(element, attr.path)
@@ -793,17 +678,25 @@ class DOMParserBase(object):
                         continue
                     attr_postprocess = attr.postprocess
                     if callable(attr_postprocess):
-                        data = attr_postprocess(data)
+                        try:
+                            data = attr_postprocess(data)
+                        except Exception, e:
+                            _m = '%s: unable to apply attr postprocess'
+                            self._logger.error(_m, self._cname, exc_info=True)
                     key = attr.key
                     if key is None:
                         key = group_key
                     elif key.startswith('.'):
                         # assuming this is an xpath
-                        key = self.xpath(element, key)[0]
+                        try:
+                            key = self.xpath(element, key)[0]
+                        except IndexError:
+                            self._logger.error('%s: XPath returned no items',
+                                                self._cname, exc_info=True)
                     elif key.startswith('self.'):
                         key = getattr(self, key[5:])
                     if attr.multi:
-                        if not result.has_key(key):
+                        if key not in result:
                             result[key] = []
                         result[key].append(data)
                     else:
